from typing import Dict, List, Tuple

from i8_terminal.common.metrics import get_all_metrics_default_period_types_dict
from i8_terminal.types.auto_complete_choice import AutoCompleteChoice

PERIOD_TYPES: Dict[str, List[Tuple[str, str]]] = {
    "D": [("p", "Default period type (daily)")],
    "R": [("p", "Default period type (realtime)")],
    "FY": [
        ("fy", "most recent fiscal year (default)"),
        ("q", "most recent quarter"),
        ("ttm", "trailing 12 months"),
        ("ytd", "year to date"),
    ],
    "Q": [
        ("q", "most recent quarter (default)"),
        ("fy", "most recent fiscal year"),
        ("ttm", "trailing 12 months"),
        ("ytd", "year to date"),
    ],
    "TTM": [
        ("ttm", "trailing 12 months (default)"),
        ("fy", "most recent fiscal year"),
        ("q", "most recent quarter"),
        ("ytd", "year to date"),
    ],
    "YTD": [
        ("ytd", "year to date (default)"),
        ("fy", "most recent fiscal year"),
        ("q", "most recent quarter"),
        ("ttm", "trailing 12 months"),
    ],
}


class PeriodParamType(AutoCompleteChoice):
    name = "period"
    metrics_default_pt: Dict[str, str] = {}

    def get_suggestions(self, keyword: str, pre_populate: bool = True, metric: str = None) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.metrics_default_pt = get_all_metrics_default_period_types_dict()
        self.set_choices(
            PERIOD_TYPES.get(self.metrics_default_pt[metric if metric else ""], [("p", "Default period type (NA)")])
        )

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]
        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "PERIOD"



from typing import Dict, List

import investor8_sdk
from investor8_sdk.models import WatchlistDto
from pandas import DataFrame
from rich.console import Console

from i8_terminal.commands.watchlist import watchlist
from i8_terminal.common.cli import pass_command
from i8_terminal.common.layout import df2Table, format_df
from i8_terminal.config import USER_SETTINGS


def get_user_watchlists() -> DataFrame:
    results = investor8_sdk.UserApi().get_watchlists_by_user(user_id=USER_SETTINGS.get("user_id"))
    return prepare_watchlists_df(results.watchlists)


def prepare_watchlists_df(watchlists: List[WatchlistDto]) -> DataFrame:
    wls: List[Dict[str, str]] = []
    for wl in watchlists:
        wls.append(
            {
                "name": wl.name,
                "tickers": f"{', '.join(str(ticker) for ticker in wl.tickers[:5])} {f'and {len(wl.tickers)-5} more' if len(wl.tickers) > 5 else ''}",  # noqa: E501
            }
        )
    watchlists_df = DataFrame(wls)
    col_names = {
        "name": "Name",
        "tickers": "Tickers",
    }
    return format_df(watchlists_df, col_names, {})


@watchlist.command()
@pass_command
def list() -> None:
    """
    Lists user watchlists.

    Examples:

    `i8 watchlist list`

    """
    console = Console()
    with console.status("Fetching data...", spinner="material"):
        df = get_user_watchlists()
    table = df2Table(df)
    console.print(table)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional


@dataclass
class ColumnInfo:
    name: str
    col_type: str
    display_name: Optional[str] = None
    data_type: Optional[str] = None
    unit: Optional[str] = None
    colorable: Optional[bool] = None

    def enrich(self, other: ColumnInfo) -> None:
        if not self.display_name:
            self.display_name = other.display_name
        if not self.data_type:
            self.data_type = other.data_type
        if not self.unit:
            self.unit = other.unit
        if not self.colorable:
            self.colorable = other.colorable



import heapq
from typing import List, Optional, Tuple

from click.types import StringParamType


class AutoCompleteChoice(StringParamType):
    name = "autocompletechoice"

    def __init__(self, choices: Optional[List[Tuple[str, str]]] = None) -> None:
        self.is_loaded = False
        self.size = 10
        if choices:
            self.set_choices(choices)

    def set_choices(self, choices: List[Tuple[str, str]]) -> None:
        if choices and len(choices) > 0:
            self._choices = choices
            self._choices_l = [(c[0].lower(), c[1].lower()) for c in choices]
            self.is_loaded = True
        else:
            self._choices = []
            self._choices_l = []

    def search_keyword(self, keyword: str) -> List[Tuple[str, str]]:
        keyword = keyword.lower()

        scores: List[Tuple[float, int]] = []
        for i, val in enumerate(self._choices_l):
            # A heurestic method to rank the list of choices
            score = 0.0
            for token in val[1].replace('"', "").split(" "):
                if token.startswith(keyword):
                    score += 1 + 1 / len(token)
            for token in val[0].replace('"', "").split("_"):
                if token.startswith(keyword):
                    score += 3 + 1 / len(val[0])
            if val[0].startswith(keyword):
                score += 5 + 1 / len(val[0])
            if val[0] == keyword:
                score += 10

            if score > 0:
                if len(scores) < self.size:
                    heapq.heappush(scores, (score, i))
                else:
                    heapq.heappushpop(scores, (score, i))

        scores_sorted = sorted(scores, key=lambda x: -x[0])
        return [self._choices[i[1]] for i in scores_sorted]

    def get_suggestions(self, keyword: str, pre_populate: bool = False) -> List[Tuple[str, str]]:
        if self._choices:
            if pre_populate and keyword.strip() == "":
                return self._choices[: self.size]
            return self.search_keyword(keyword)
        return []

    def __repr__(self) -> str:
        return "AUTOCOMPLETECHOICE"



from typing import Any, Dict, Optional

import click
import investor8_sdk
import pandas as pd
from rich.console import Console

from i8_terminal.commands.watchlist import watchlist
from i8_terminal.common.cli import pass_command
from i8_terminal.common.layout import df2Table
from i8_terminal.common.metrics import (
    get_current_metrics_df,
    get_view_metrics,
    prepare_current_metrics_formatted_df,
)
from i8_terminal.common.utils import export_data, export_to_html
from i8_terminal.config import APP_SETTINGS, USER_SETTINGS
from i8_terminal.types.metric_param_type import MetricParamType
from i8_terminal.types.metric_view_param_type import MetricViewParamType
from i8_terminal.types.user_watchlists_param_type import UserWatchlistsParamType


def prepare_watchlist_stocks_df(name: str, metrics: str) -> Optional[pd.DataFrame]:
    watchlist = investor8_sdk.UserApi().get_watchlist_by_name_user_id(name=name, user_id=USER_SETTINGS.get("user_id"))
    watchlist_stocks_df = get_current_metrics_df(",".join(watchlist.tickers), metrics)
    return watchlist_stocks_df


@watchlist.command()
@click.option(
    "--name",
    "-n",
    type=UserWatchlistsParamType(),
    required=True,
    help="Name of the watchlist.",
)
@click.option(
    "--metrics",
    "-m",
    type=MetricParamType(),
    help="Comma-separated list of daily metrics.",
)
@click.option("--export", "export_path", "-e", help="Filename to export the output to.")
@click.option(
    "--view_name", "view_name", "-v", type=MetricViewParamType(), help="Metric view name in configuration file."
)
@pass_command
def metrics(name: str, metrics: str, export_path: Optional[str], view_name: Optional[str]) -> None:
    """
    Lists and compares watchlist companies based on a given list of metrics.

    Examples:

    `i8 watchlist metrics --name MyWatchlist --metrics total_revenue,net_income,price_to_earnings`

    """
    console = Console()
    if not metrics and not view_name:
        console.print("The 'metrics' or 'view_name' parameter must be provided", style="yellow")
        return
    if view_name and metrics:
        console.print("The 'metrics' or 'view_name' options are mutually exclusive", style="yellow")
        return
    if view_name:
        metrics = ",".join(get_view_metrics(view_name))
    with console.status("Fetching data...", spinner="material"):
        df = prepare_watchlist_stocks_df(name, metrics)
    if df is None:
        console.print("No data found for metrics with selected tickers", style="yellow")
        return
    for m in [*set(metric.split(".")[0] for metric in set(metrics.split(","))) - set(df["metric_name"])]:
        console.print(f"\nNo data found for metric {m} with selected tickers", style="yellow")
    columns_justify: Dict[str, Any] = {}
    if export_path:
        if export_path.split(".")[-1] == "html":
            for metric_display_name, metric_df in df.groupby("display_name"):
                columns_justify[metric_display_name] = (
                    "left" if metric_df["display_format"].values[0] == "str" else "right"
                )
            table = df2Table(prepare_current_metrics_formatted_df(df, "console"), columns_justify=columns_justify)
            export_to_html(table, export_path)
            return
        export_data(
            prepare_current_metrics_formatted_df(df, "store"),
            export_path,
            column_width=18,
            column_format=APP_SETTINGS["styles"]["xlsx"]["financials"]["column"],
        )
    else:
        for metric_display_name, metric_df in df.groupby("display_name"):
            columns_justify[metric_display_name] = "left" if metric_df["display_format"].values[0] == "str" else "right"
        table = df2Table(prepare_current_metrics_formatted_df(df, "console"), columns_justify=columns_justify)
        console.print(table)



from typing import Any, Dict, List, Optional, Tuple

import click
import investor8_sdk
import pandas as pd
from rich.console import Console

from i8_terminal.commands.screen import screen
from i8_terminal.common.cli import pass_command
from i8_terminal.common.layout import df2Table
from i8_terminal.common.metrics import (
    get_current_metrics_df,
    get_metric_info,
    get_view_metrics,
    prepare_current_metrics_formatted_df,
)
from i8_terminal.common.utils import export_data, export_to_html
from i8_terminal.config import APP_SETTINGS
from i8_terminal.types.metric_identifier_param_type import MetricIdentifierParamType
from i8_terminal.types.metric_view_param_type import MetricViewParamType
from i8_terminal.types.screening_condition_param_type import ScreeningConditionParamType
from i8_terminal.types.sort_order_param_type import SortOrderParamType

RELATIVE_PERIOD_TYPES: Dict[str, str] = {
    "D": ".d",
    "R": ".r",
    "FY": ".fy",
    "Q": ".q",
}


def prepare_screen_df(
    conditions: List[str], metrics: str, sort_by: Optional[str], sort_order: Optional[str]
) -> Tuple[List[str], Optional[pd.DataFrame]]:
    max_count = 20
    for index, condition in enumerate(conditions):
        condition_parts = condition.split(":")
        metric = condition_parts[0]
        metric_parts = metric.split(".")
        if len(metric_parts) == 1:
            metric_default_period_type = get_metric_info(metric_parts[0])["default_period_type"]
            period_type = RELATIVE_PERIOD_TYPES.get(metric_default_period_type, "")
            metric_new = f"{metric}{period_type}"
            conditions[index] = conditions[index].replace(metric, metric_new)
    if not sort_by:
        sort_by = metrics.split(",")[0]
    tickers_list = investor8_sdk.ScreenerApi().search(
        conditions=",".join(conditions), order_by=sort_by, order_direction=sort_order
    )
    screen_df = get_current_metrics_df(",".join(tickers_list[:max_count]), metrics)
    return tickers_list, screen_df


def sort_by_tickers(df: pd.DataFrame, sorted_tickers: List[str]) -> pd.DataFrame:
    sorterIndex = dict(zip(sorted_tickers, range(len(sorted_tickers))))
    df["Rank"] = df["Ticker"].map(sorterIndex)
    df.sort_values("Rank", ascending=True, inplace=True)
    df.drop("Rank", axis=1, inplace=True)
    return df


@screen.command()
@click.option(
    "--condition",
    "condition",
    "-c",
    required=True,
    type=ScreeningConditionParamType(),
    multiple=True,
    help="Condition of metric.",
)
@click.option(
    "--view_name", "view_name", "-v", type=MetricViewParamType(), help="Metric view name in configuration file."
)
@click.option(
    "--metrics",
    "metrics",
    "-m",
    type=MetricIdentifierParamType(),
    help="Comma-separated list of daily metrics.",
)
@click.option("--export", "export_path", "-e", help="Filename to export the output to.")
@click.option(
    "--sort_by",
    "sort_by",
    "-sb",
    type=MetricIdentifierParamType(),
    help="Metric to sort the output by.",
)
@click.option(
    "--sort_order", "sort_order", "-so", type=SortOrderParamType(), help="Order to sort the output by.", default="desc"
)
@pass_command
def search(
    condition: Tuple[str],
    view_name: Optional[str],
    metrics: Optional[str],
    export_path: Optional[str],
    sort_by: Optional[str],
    sort_order: Optional[str],
) -> None:
    console = Console()
    if not metrics and not view_name:
        console.print("The 'metrics' or 'view_name' parameter must be provided", style="yellow")
        return
    if view_name and metrics:
        console.print(
            "The 'metrics' or 'view_name' options are mutually exclusive. Provide a value only for one of them.",
            style="yellow",
        )
        return
    if view_name:
        metrics = ",".join(get_view_metrics(view_name))
    with console.status("Fetching data...", spinner="material"):
        sorted_tickers, df = prepare_screen_df(list(condition), metrics, sort_by, sort_order)  # type: ignore
    if df is None:
        console.print("No data found for the provided screen conditions", style="yellow")
        return
    no_data_metrics = [
        *set(metric.split(".")[0] for metric in set(metrics.split(","))) - set(df["metric_name"])  # type: ignore
    ]
    for m in no_data_metrics:
        console.print(f"\nNo data found for metric {m} with selected tickers", style="yellow")
    columns_justify: Dict[str, Any] = {}
    if export_path:
        if export_path.split(".")[-1] == "html":
            for metric_display_name, metric_df in df.groupby("display_name"):
                columns_justify[metric_display_name] = (
                    "left" if metric_df["display_format"].values[0] == "str" else "right"
                )
            table = df2Table(
                sort_by_tickers(prepare_current_metrics_formatted_df(df, "console"), sorted_tickers),
                columns_justify=columns_justify,
            )
            export_to_html(table, export_path)
            return
        export_data(
            sort_by_tickers(prepare_current_metrics_formatted_df(df, "store"), sorted_tickers),
            export_path,
            column_width=18,
            column_format=APP_SETTINGS["styles"]["xlsx"]["financials"]["column"],
        )
    else:
        for metric_display_name, metric_df in df.groupby("display_name"):
            columns_justify[metric_display_name] = "left" if metric_df["display_format"].values[0] == "str" else "right"
        table = df2Table(
            sort_by_tickers(prepare_current_metrics_formatted_df(df, "console"), sorted_tickers),
            columns_justify=columns_justify,
        )
        console.print(table)



from typing import List, Tuple

from i8_terminal.types.auto_complete_choice import AutoCompleteChoice
from i8_terminal.types.condition_param_type import ConditionParamType
from i8_terminal.types.metric_param_type import MetricParamType
from i8_terminal.types.period_param_type import PeriodParamType
from i8_terminal.types.screening_operator_param_type import ScreeningOperatorParamType
from i8_terminal.types.screening_value_field_param_type import (
    ScreeningValueFieldParamType,
)


class ScreeningConditionParamType(AutoCompleteChoice):
    name = "screeningcondition"

    def __init__(self) -> None:
        self._metric_auto_comp = MetricParamType()
        self._period_auto_comp = PeriodParamType()
        self._value_field_auto_comp = ScreeningValueFieldParamType()
        self._operator_auto_comp = ScreeningOperatorParamType()
        self._condition_auto_comp = ConditionParamType()

    def get_suggestions(
        self,
        keyword: str,
        pre_populate: bool = True,
        param_type: str = None,
        metric: str = None,
        period: str = None,
        value_field: str = None,
    ) -> List[Tuple[str, str]]:
        if param_type == "metric":
            return self._metric_auto_comp.get_suggestions(keyword)
        elif param_type == "period":
            return self._period_auto_comp.get_suggestions(keyword, metric=metric)  # type: ignore
        elif param_type == "value_field":
            return self._value_field_auto_comp.get_suggestions(keyword, metric=metric)  # type: ignore
        elif param_type == "operator":
            return self._operator_auto_comp.get_suggestions(keyword, metric=metric)  # type: ignore
        else:
            return self._condition_auto_comp.get_suggestions(
                keyword, metric=metric, period=period, value_field=value_field
            )  # type: ignore

    def __repr__(self) -> str:
        return "SCREENINGCONDITION"



from typing import Any, Dict, Optional

import investor8_sdk
import pandas as pd
from rich.table import Table

from i8_terminal.common.layout import df2Table
from i8_terminal.common.metrics import (
    get_current_metrics_df,
    get_view_metrics,
    prepare_current_metrics_formatted_df,
)
from i8_terminal.common.utils import export_data, export_to_html
from i8_terminal.config import APP_SETTINGS


def get_top_stocks_df(category: str, index: str, view_name: Optional[str]) -> Optional[pd.DataFrame]:
    metrics = APP_SETTINGS["commands"]["screen_gainers"]["metrics"]
    companies_data = investor8_sdk.ScreenerApi().get_top_stocks(category, index=index)
    if companies_data is None:
        return None
    companies = [company.ticker for company in companies_data]
    if view_name:
        metrics = metrics + "," + ",".join(get_view_metrics(view_name))
    return get_current_metrics_df(",".join(companies), metrics)


def render_top_stocks(df: pd.DataFrame, export_path: Optional[str], ascending: bool = True) -> Optional[Table]:
    columns_justify: Dict[str, Any] = {}
    for metric_display_name, metric_df in df.groupby("display_name"):
        columns_justify[metric_display_name] = "left" if metric_df["display_format"].values[0] == "str" else "right"
    change_rows = df.loc[df["metric_name"] == "change"]
    df = pd.concat(
        [
            pd.DataFrame(change_rows.replace({"change": "change_numeric", "Change": "Change Numeric", "perc": "str"})),
            df,
        ],
        ignore_index=True,
        axis=0,
    )
    if export_path:
        if export_path.split(".")[-1] == "html":
            formatted_df = prepare_current_metrics_formatted_df(df, "console").sort_values(
                "Change Numeric", ascending=ascending
            )
            formatted_df.drop("Change Numeric", axis=1, inplace=True)
            table = df2Table(
                formatted_df,
                columns_justify=columns_justify,
            )
            export_to_html(table, export_path)
            return None
        formatted_df = prepare_current_metrics_formatted_df(df, "store").sort_values(
            "Change Numeric", ascending=ascending
        )
        formatted_df.drop("Change Numeric", axis=1, inplace=True)
        export_data(
            formatted_df,
            export_path,
            column_width=18,
            column_format=APP_SETTINGS["styles"]["xlsx"]["financials"]["column"],
        )
        return None
    else:
        formatted_df = prepare_current_metrics_formatted_df(df, "console").sort_values(
            "Change Numeric", ascending=ascending
        )
        formatted_df.drop("Change Numeric", axis=1, inplace=True)
        table = df2Table(
            formatted_df,
            columns_justify=columns_justify,
        )
        return table



from typing import Dict, List, Tuple

from i8_terminal.common.metrics import get_all_metrics_df
from i8_terminal.types.auto_complete_choice import AutoCompleteChoice

SCREENING_OPERATORS: Dict[str, List[Tuple[str, str]]] = {
    "Numeric": [
        ("gt", "greater than"),
        ("lt", "less than"),
        ("bw", "between"),
    ],
    "Non-Numeric": [
        ("eq", "equal"),
    ],
}


def get_metrics_data_format_dict() -> Dict[str, str]:
    df = get_all_metrics_df()[["metric_name", "data_format"]]
    return dict([(i, j) for i, j in zip(df.metric_name, df.data_format)])


class ScreeningOperatorParamType(AutoCompleteChoice):
    name = "screeningoperator"
    metrics_default_pt: Dict[str, str] = {}

    def get_suggestions(self, keyword: str, pre_populate: bool = True, metric: str = None) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.metrics_data_formats = get_metrics_data_format_dict()
        self.set_choices(
            SCREENING_OPERATORS.get(
                "Numeric"
                if self.metrics_data_formats.get(metric, "str") in ["float", "int", "unsigned_int"]  # type: ignore
                else "Non-Numeric"
            )
        )  # type: ignore

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]
        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "SCREENINGOPERATOR"



from typing import List

import click
import investor8_sdk
from rich.console import Console
from rich.style import Style

from i8_terminal.app.layout import get_terminal_command_layout
from i8_terminal.commands.watchlist import watchlist
from i8_terminal.common.cli import pass_command
from i8_terminal.common.stock_info import validate_tickers
from i8_terminal.config import USER_SETTINGS
from i8_terminal.types.user_watchlist_tickers_param_type import (
    UserWatchlistTickersParamType,
)
from i8_terminal.types.user_watchlists_param_type import UserWatchlistsParamType


def remove_tickers_from_watchlist(name: str, tickers: List[str]) -> None:
    wl = investor8_sdk.UserApi().get_watchlist_by_name_user_id(name=name, user_id=USER_SETTINGS.get("user_id"))
    for ticker in tickers:
        investor8_sdk.UserApi().remove_from_watchlist(body={"Id": wl.id, "Ticker": ticker})


@watchlist.command()
@click.option(
    "--name",
    "-n",
    type=UserWatchlistsParamType(),
    required=True,
    help="Name of the watchlist.",
)
@click.option(
    "--tickers",
    "-k",
    type=UserWatchlistTickersParamType(),
    required=True,
    callback=validate_tickers,
    help="Comma-separated list of tickers.",
)
@pass_command
def rm(name: str, tickers: str) -> None:
    """
    Removes given tickers from a given watchlist.

    Examples:

    `i8 watchlist rm --name MyWatchlist --tickers AAPL,MSFT`

    """
    console = Console()
    tickers_list = tickers.replace(" ", "").upper().split(",")
    with console.status("Updating Watchlist...", spinner="material"):
        remove_tickers_from_watchlist(name, tickers_list)
    console.print(
        f"✅ Ticker{'s' if len(tickers_list) > 1 else ''} [cyan]{', '.join(tickers_list)}[/cyan] removed from watchlist [cyan]{name}[/cyan] successfully!"  # noqa: E501
    )
    terminal_command_style = Style(**get_terminal_command_layout())
    console.print(
        f'Try `[{terminal_command_style}]watchlist summary --name "{name}"[/{terminal_command_style}]`to see the watchlist.'  # noqa: E501
    )



import click
import investor8_sdk
import pandas as pd
from rich.console import Console

from i8_terminal.commands.watchlist import watchlist
from i8_terminal.common.cli import pass_command
from i8_terminal.common.metrics import (
    get_current_metrics_df,
    prepare_current_metrics_formatted_df,
)
from i8_terminal.config import APP_SETTINGS, USER_SETTINGS
from i8_terminal.types.user_watchlists_param_type import UserWatchlistsParamType


def export_watchlist_data(
    name: str,
    path: str,
) -> None:
    console = Console()
    extension = path.split(".")[-1]
    if extension != "xlsx":
        console.print("\n⚠ Error: path is not valid", style="yellow")
        return
    writer = pd.ExcelWriter(path, engine="xlsxwriter")
    workbook = writer.book
    column_width = 18
    header_format = workbook.add_format(APP_SETTINGS["styles"]["xlsx"]["default"]["header"])
    metric_format = workbook.add_format(APP_SETTINGS["styles"]["xlsx"]["default"]["metric"])
    column_format = workbook.add_format(APP_SETTINGS["styles"]["xlsx"]["company"]["column"])
    tickers = (
        investor8_sdk.UserApi().get_watchlist_by_name_user_id(name=name, user_id=USER_SETTINGS.get("user_id")).tickers
    )
    summary_df = prepare_current_metrics_formatted_df(
        get_current_metrics_df(
            ",".join(tickers), "company_name,stock_exchange,price,change,52_week_low,52_week_high,marketcap"
        ),
        "store",
    )
    summary_df.to_excel(writer, sheet_name="Summary", startrow=1, header=False, index=False)
    worksheet = writer.sheets["Summary"]
    headers = summary_df.columns.tolist()
    for col_num, value in enumerate(headers):
        if type(value) is tuple:
            worksheet.merge_range(0, col_num, len(value) - 1, col_num, " ".join(reversed(value)), header_format)
        else:
            worksheet.write(0, col_num, value, header_format)
    worksheet.set_column(0, 0, 20, metric_format)
    worksheet.set_column(1, len(summary_df.columns) - 1, column_width, column_format)
    financials_df = prepare_current_metrics_formatted_df(
        get_current_metrics_df(
            ",".join(tickers),
            "total_revenue,net_income,basic_eps,net_cash_from_operating_activities,total_assets,total_liabilities",
        ),
        "store",
    )
    financials_df.to_excel(writer, sheet_name="Financials", startrow=1, header=False, index=False)
    worksheet = writer.sheets["Financials"]
    headers = financials_df.columns.tolist()
    for col_num, value in enumerate(headers):
        if type(value) is tuple:
            worksheet.merge_range(0, col_num, len(value) - 1, col_num, " ".join(reversed(value)), header_format)
        else:
            worksheet.write(0, col_num, value, header_format)
    worksheet.set_column(0, 0, 20, metric_format)
    worksheet.set_column(1, len(financials_df.columns) - 1, column_width, column_format)
    writer.save()
    console.print(f"\nData is saved on: {path}")


@watchlist.command()
@click.option(
    "--name",
    "-n",
    type=UserWatchlistsParamType(),
    required=True,
    help="Name of the watchlist.",
)
@click.option("--path", "path", "-p", required=True, help="Filename to export the output to.")
@pass_command
def export(name: str, path: str) -> None:
    """
    Exports a given watchlist to an excel file.

    Examples:

    `i8 watchlist export --name MyWatchlist --path MyWatchlist.xlsx`

    """
    console = Console()
    with console.status("Fetching data...", spinner="material"):
        export_watchlist_data(name, path)



import os
import sys

from rich.console import Console

from i8_terminal.common.cli import log_terminal_usage, pass_command
from i8_terminal.config import USER_SETTINGS, init_api_configs, is_user_logged_in
from i8_terminal.utils_setup import get_version

console = Console(force_terminal=True, color_system="truecolor")
if set(["i8"]) == set(sys.argv[1:]):
    console.print(f"\n👋 Welcome to i8 Terminal! Version: {get_version()}", style="yellow")
    console.print("Copyright © 2020-2022 Investoreight | https://www.i8terminal.io/\n")
    console.print("- Enter [magenta]i8 shell[/magenta] to run i8 Terminal shell.")
    os._exit(0)
if set(["i8", "version"]) == set(sys.argv[1:]):
    console.print(f"Version: {get_version()}", style="yellow")
    os._exit(0)
status = console.status("Starting Up...", spinner="material")
status.start()


import webbrowser

import click
import investor8_sdk
from click_repl import repl
from investor8_sdk.rest import ApiException

from i8_terminal.commands import cli
from i8_terminal.common.stock_info import validate_ticker
from i8_terminal.types.i8_auto_suggest import I8AutoSuggest
from i8_terminal.types.i8_completer import I8Completer
from i8_terminal.types.ticker_param_type import TickerParamType


def init_commands() -> None:
    status.start()
    app_dir = os.path.join(os.path.join(os.path.dirname(sys.executable), "lib"), "i8_terminal")
    sys.path.append(app_dir)
    commands_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "commands")
    commands_dir = commands_dir if os.path.exists(commands_dir) else os.path.join(app_dir, "commands")
    ignore_dir = ["__pycache__"]
    for cmd in [p for p in os.listdir(commands_dir) if os.path.isdir(os.path.join(commands_dir, p))]:
        if cmd not in ignore_dir:
            for sub_cmd in os.listdir(os.path.join(commands_dir, cmd)):
                sub_cmd_splitted = sub_cmd.split(".")
                if (sub_cmd_splitted[-1] in ["py", "pyc"]) and sub_cmd_splitted[0] not in ["__init__"]:
                    __import__(f"i8_terminal.commands.{cmd}.{''.join(sub_cmd_splitted[:-1])}")
    status.stop()

    @cli.command()
    def shell() -> None:
        """Open i8-shell."""
        print_welcome_msg()
        prompt_kwargs = {"completer": I8Completer(cli), "auto_suggest": I8AutoSuggest(cli)}

        while True:
            try:
                repl(click.get_current_context(), prompt_kwargs=prompt_kwargs)
            except ApiException as e:
                if "apiKey" in e.body.decode("utf-8"):
                    console.print(
                        "You need to login before using i8 Terminal. Please login to i8 Terminal using [magenta]user login[/magenta] command."  # noqa: E501
                    )
                else:
                    console.print(f"⚠ Error: {e.body.decode('utf-8')}", style="yellow")
                log_terminal_usage(click.get_current_context(), e.body.decode("utf-8"))
            except Exception as e:
                display_error = f"- Type: {type(e).__name__}\n- Message: {e}"
                log_terminal_usage(click.get_current_context(), display_error)
                console.print(f"⚠ Error:\n{display_error}", style="yellow")

    @cli.command()
    def exit() -> None:
        """Exit the terminal."""
        os._exit(0)

    @cli.command()
    @pass_command
    def version() -> None:
        """Get i8-terminal version."""
        click.echo(get_version())

    @cli.command()
    @click.option("--all", "-a", is_flag=True, default=False, help="Clear entire screen.")
    @pass_command
    def clear(all: bool) -> None:
        """Clear the console screen."""
        cls_screen()
        if not all:
            print_welcome_msg()

    @cli.command()
    @click.option(
        "--ticker",
        "-k",
        type=TickerParamType(),
        required=True,
        callback=validate_ticker,
        help="Ticker or company name.",
    )
    @pass_command
    def browse(ticker: str) -> None:
        """
        Open company detail page in investoreight.com

        Examples:

        `i8 browse --ticker MSFT`
        """
        url = f"https://www.investoreight.com/stock/{ticker}"
        webbrowser.open(url)
        console.print(f"[blue]{url}[/blue]")


def print_welcome_msg() -> None:
    console.print(f"\n👋 Welcome to i8 Terminal! Version:  {get_version()}", style="yellow")
    console.print("Copyright © 2020-2022 Investoreight | https://www.i8terminal.io/")
    console.print("- Enter [magenta]?[/magenta] to get the list of commands.")
    console.print("- Enter [magenta]exit[/magenta] to exit the shell.\n")
    if is_user_logged_in():
        console.print(f'Logged in as "{USER_SETTINGS["user_id"]}".\n')
    else:
        console.print(
            "You are not logged in. Please login to i8 Terminal using [magenta]user login[/magenta] command.\n"
        )


def cls_screen() -> None:
    os.system("cls" if os.name == "nt" else "clear")


def check_version() -> None:
    resp = None
    try:
        resp = investor8_sdk.SettingsApi().check_i8t_version(get_version())
    except Exception:
        pass
    if not resp or not resp.to_dict().get("version_supported"):
        status.stop()
        console.print(
            "[yellow]You are using an old version of i8 Terminal that is not supported anymore.[/yellow]",
            "[yellow]Please update i8 Terminal with the following command to be able to use the application.[/yellow]",
            "\n[magenta]i8update[/magenta]\n",
            "If you are using Python pip, you can run the following command to update i8 Terminal:\n",
            "[magenta]pip install --upgrade i8-terminal[/magenta]",
        )
        os._exit(0)


def main() -> None:
    check_version()
    init_commands()
    if is_user_logged_in():
        init_api_configs()

    try:
        cli(obj={})
    except ApiException as e:
        if "apiKey" in e.body.decode("utf-8"):
            console.print(
                "You need to login before using i8 Terminal. Please login to i8 Terminal using [magenta]user login[/magenta] command."  # noqa: E501
            )
        else:
            console.print(f"⚠ Error: {e.body.decode('utf-8')}", style="yellow")
    except Exception as e:
        display_error = f"- Type: {type(e).__name__}\n- Message: {e}"
        console.print(f"⚠ Error:\n{display_error}", style="yellow")


if __name__ == "__main__":
    main()
else:
    init_commands()



from typing import List, Tuple

import investor8_sdk
import pandas as pd

from i8_terminal.types.auto_complete_choice import AutoCompleteChoice


def get_metric_view_names() -> List[Tuple[str, str]]:
    results = investor8_sdk.MetricsApi().get_list_metric_views()
    df = pd.DataFrame([d.to_dict() for d in results])[["view_name", "display_name"]]
    return list(df.to_records(index=False))


class MetricViewParamType(AutoCompleteChoice):
    name = "metricview"

    def get_suggestions(self, keyword: str, pre_populate: bool = True) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.set_choices(get_metric_view_names())

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]
        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "METRICVIEW"



import contextlib
import logging
import os
import socket
import sys
import webbrowser
from threading import Timer
from typing import Any, Dict, List, Tuple

import dash
import dash_bootstrap_components as dbc
import investor8_sdk
import plotly.graph_objects as go
from dash import html
from dash.dependencies import Input, Output, State
from flask import request
from rich.console import Console

from i8_terminal import config
from i8_terminal.app.layout import create_plot_layout
from i8_terminal.common.formatting import make_svg_responsive
from i8_terminal.config import APP_SETTINGS, ASSETS_PATH, USER_SETTINGS

APP = dash.Dash(external_stylesheets=[dbc.themes.BOOTSTRAP], assets_folder=ASSETS_PATH)


@APP.callback(
    Output("savePlotModal", "is_open"),
    Input("openModalBtn", "n_clicks"),
    Input("saveBtn", "n_clicks"),
    State("savePlotModal", "is_open"),
)
def toggle_modal(open_btn_clicks: int, save_btn_clicks: int, is_open: bool) -> bool:
    if not open_btn_clicks:
        return False
    return not is_open


@APP.callback(
    Output("fakeOutput", "children"),
    Input("closeServerBtn", "n_clicks"),
)
def shutdown_server(close_btn_clicks: int) -> str:
    if close_btn_clicks:
        shutdown_func = request.environ.get("werkzeug.server.shutdown")
        if shutdown_func is None:
            raise RuntimeError("Not running werkzeug")
        shutdown_func()
    return ""


@APP.callback(
    Output("saveAlert", "is_open"),
    Output("saveAlert", "children"),
    Input("saveBtn", "n_clicks"),
    State("cmdContextStore", "data"),
    State("titleInput", "value"),
    State("userNotesInput", "value"),
    State("isPublicBtn", "value"),
    State("figDict", "data"),
)
def save_button(
    save_btn_clicks: int,
    cmd_context: Dict[str, Any],
    title: str,
    user_notes: str,
    is_public: List[int],
    fig_dict: Dict[str, Any],
) -> Tuple[bool, Any]:
    if save_btn_clicks:
        fig_obj = go.Figure(fig_dict)
        fig_obj_small = go.Figure(fig_obj)
        fig_obj_small.update_layout(
            title=dict(text="", font=dict(size=14)),
            font=dict(size=9),
            legend_title=dict(font=dict(size=11)),
            legend=dict(font=dict(size=9)),
        )
        thumbnail = make_svg_responsive(fig_obj_small.to_image(format="svg", width=846, height=500).decode("utf-8"))
        tickers = cmd_context["tickers"]
        fig_obj.layout.images[0].source = config.I8_TERMINAL_LOGO_URL  # Update plot logo to url
        fig_obj.layout.title.text = title  # Update plot title
        body = {
            "Title": title,
            "Tickers": tickers if type(tickers) is list else [tickers],
            "UserId": USER_SETTINGS["user_id"],
            "PlotData": fig_obj.to_json(),
            "IsPublic": 1 in is_public,
            "I8Command": cmd_context["command_path"],
            "UserNotes": user_notes,
            "PlotType": cmd_context["plot_type"],
            "Thumbnail": thumbnail,
            # TODO: implement tags
            "Tags": [],
        }
        resp = investor8_sdk.UserApi().create_plot(body=body)
        plot_url = f"https://www.investoreight.com/plot/{resp.id}"
        alert_content = html.Div(["Your plot is saved and published at: ", html.A(plot_url, href=plot_url)])

        return True, alert_content

    return False, None


@APP.callback(
    Output("loadingGif", "hidden"),
    Input("saveBtn", "n_clicks"),
    State("loadingGif", "hidden"),
)
def show_loading(save_btn_clicks: int, hidden: bool) -> bool:
    if not save_btn_clicks:
        return True
    return not hidden


@APP.callback(
    Output("loadingGif", "style"),
    Input("saveAlert", "is_open"),
)
def hide_loading(is_open: bool) -> Dict[str, Any]:
    if is_open:
        return {"display": "none"}
    return {
        "position": "fixed",
        "width": "100%",
        "height": "100%",
        "left": 0,
        "right": 0,
        "bottom": 0,
        "background-color": "rgba(0,0,0,0.5)",
        "z-index": 2,
    }


def _configure_dash() -> None:
    # TODO: the following method does not disable dash logging (it should)
    dash_logger = logging.getLogger("dash")
    dash_logger.setLevel(logging.WARNING)
    dash_logger.disabled = True

    cli = sys.modules["flask.cli"]
    cli.show_server_banner = lambda *x: None  # type: ignore


def serve_plot(fig: go.Figure, cmd_context: Dict[str, Any]) -> None:
    _configure_dash()

    APP.layout = create_plot_layout(fig, cmd_context)
    APP.title = f"i8 Terminal: {cmd_context['plot_title']}"
    app_url = f"http://localhost:{APP_SETTINGS['app']['port']}/"
    console = Console()
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        is_port_in_use = s.connect_ex(("localhost", APP_SETTINGS["app"]["port"])) == 0
    if is_port_in_use:
        console.print(
            f"Port number {APP_SETTINGS['app']['port']} is already in use. Please make sure that only one instance of i8-terminal runs at a time!",  # noqa: E501
            style="yellow",
        )
        return
    with open(os.devnull, "w") as f, contextlib.redirect_stderr(f):
        Timer(2, lambda: webbrowser.open_new(app_url)).start()
        console.print(
            f'[bold green]Your plot is serving on http://127.0.0.1:{APP_SETTINGS["app"]["port"]}/[/bold green]'
        )
        console.print("Press `Ctrl + C` to stop the webserver.")
        if APP.logger.hasHandlers():
            APP.logger.handlers.clear()
        APP.run_server(debug=APP_SETTINGS["app"]["debug"])
